@@ -11,7 +11,7 @@ import numpy as np
 
 
 class AggregateCallback:
-    def __init__(self, handler, *args, **kwargs):
+    def __init__(self, handler):
         self.handler = handler
 
 
@@ -21,16 +21,16 @@ class Throttle(AggregateCallback):
     1 update per `window` interval; all others are dropped
     """
 
-    def __init__(self, *args, window=60, **kwargs):
-        super().__init__(*args, **kwargs)
+    def __init__(self, handler, window=60):
+        super().__init__(handler)
         self.window = window
         self.last_update = 0
 
-    async def __call__(self, **kwargs):
+    async def __call__(self, data):
         now = time.time()
         if now - self.last_update > self.window:
             self.last_update = now
-            await self.handler(**kwargs)
+            await self.handler(data)
 
 
 class OHLCV(AggregateCallback):
@@ -41,8 +41,8 @@ class OHLCV(AggregateCallback):
     You should probably use the candle data channel (if the exchange supports that).
     """
 
-    def __init__(self, *args, window=300, **kwargs):
-        super().__init__(*args, **kwargs)
+    def __init__(self, *args, window=300):
+        super().__init__(*args)
         self.window = window
         self.last_update = time.time()
         self.data = {}
@@ -60,16 +60,17 @@ class OHLCV(AggregateCallback):
             self.data[symbol]['low'] = price
         self.data[symbol]['vwap'] += price * amount
 
-    async def __call__(self, *, feed: str, symbol: str, side: str, amount: Decimal, price: Decimal, order_id=None, timestamp: float, receipt_timestamp: float, order_type=None):
+    async def __call__(self, trade, receipt_timestamp: float):
         now = time.time()
         if now - self.last_update > self.window:
             self.last_update = now
             for p in self.data:
                 self.data[p]['vwap'] /= self.data[p]['volume']
-            await self.handler(data=self.data)
+
+            await self.handler(self.data)
             self.data = {}
 
-        self._agg(symbol, amount, price)
+        self._agg(trade.symbol, trade.amount, trade.price)
 
 
 class RenkoFixed(AggregateCallback):
@@ -130,10 +131,10 @@ class RenkoFixed(AggregateCallback):
         else:
             self.new_brick = False
 
-    async def __call__(self, *, feed: str, symbol: str, side: str, amount: Decimal, price: Decimal, order_id=None, timestamp: float, receipt_timestamp: float, order_type=None):
+    async def __call__(self, trade, receipt_timestamp: float):
         if self.new_brick:
-            await self.handler(data=self.data)
-        self._agg(symbol, price)
+            await self.handler(self.data)
+        self._agg(trade.symbol, trade.price)
 
 
 class CustomAggregate(AggregateCallback):
@@ -153,11 +154,11 @@ class CustomAggregate(AggregateCallback):
         self.data = {}
         self.init(self.data)
 
-    async def __call__(self, **kwargs):
+    async def __call__(self, dtype, receipt_timestamp: float):
         now = time.time()
         if now - self.last_update > self.window:
             self.last_update = now
-            await self.handler(data=self.data)
+            await self.handler(self.data)
             self.init(self.data)
 
-        self.agg(self.data, **kwargs)
+        self.agg(self.data, dtype, receipt_timestamp)
