@@ -51,34 +51,35 @@ UPBIT = 'UPBIT'
 L1_BOOK = 'l1_book'
 L2_BOOK = 'l2_book'
 L3_BOOK = 'l3_book'
-BOOK_DELTA = 'book_delta'
 TRADES = 'trades'
 TICKER = 'ticker'
 FUNDING = 'funding'
 OPEN_INTEREST = 'open_interest'
 LIQUIDATIONS = 'liquidations'
-FUTURES_INDEX = 'futures_index'
+INDEX = 'index'
 UNSUPPORTED = 'unsupported'
 CANDLES = 'candles'
-LAST_PRICE = 'last_price'
 
 # Account Data / Authenticated Channels
 ORDER_INFO = 'order_info'
-USER_FILLS = 'user_fills'
+FILLS = 'fills'
 TRANSACTIONS = 'transactions'
 BALANCES = 'balances'
+POSITIONS = 'positions'
 PLACE_ORDER = 'place_order'
 CANCEL_ORDER = 'cancel_order'
 ORDERS = 'orders'
 ORDER_STATUS = 'order_status'
 TRADE_HISTORY = 'trade_history'
-USER_DATA = 'user_data'
+POSITIONS = 'positions'
 
 BUY = 'buy'
 SELL = 'sell'
 BID = 'bid'
 ASK = 'ask'
 UND = 'undefined'
+MAKER = 'maker'
+TAKER = 'taker'
 
 LIMIT = 'limit'
 MARKET = 'market'
@@ -87,6 +88,7 @@ STOP_MARKET = 'stop-market'
 MAKER_OR_CANCEL = 'maker-or-cancel'
 FILL_OR_KILL = 'fill-or-kill'
 IMMEDIATE_OR_CANCEL = 'immediate-or-cancel'
+GOOD_TIL_CANCELED = 'good-til-canceled'
 
 OPEN = 'open'
 PENDING = 'pending'
@@ -96,6 +98,10 @@ CANCELLED = 'cancelled'
 UNFILLED = 'unfilled'
 EXPIRED = 'expired'
 SUSPENDED = 'suspended'
+FAILED = 'failed'
+SUBMITTING = 'submitting'
+CANCELLING = 'cancelling'
+CLOSED = 'closed'
 
 # Instrument Definitions
 
