@@ -5,25 +5,31 @@ Please see the LICENSE file for the terms and conditions
 associated with this software.
 '''
 import logging
-from asyncio import create_task
+from asyncio import create_task, sleep
 from collections import defaultdict, deque
 from decimal import Decimal
+import requests
+import time
 from typing import Dict, Union, Tuple
+from urllib.parse import urlencode
 
-from sortedcontainers import SortedDict as sd
 from yapic import json
 
-from cryptofeed.connection import AsyncConnection, HTTPPoll
-from cryptofeed.defines import BID, ASK, BINANCE, BUY, CANDLES, FUNDING, FUTURES, L2_BOOK, LIQUIDATIONS, OPEN_INTEREST, PERPETUAL, SELL, SPOT, TICKER, TRADES, FILLED, UNFILLED
+from cryptofeed.connection import AsyncConnection, HTTPPoll, HTTPConcurrentPoll
+from cryptofeed.defines import ASK, BALANCES, BID, BINANCE, BUY, CANDLES, FUNDING, FUTURES, L2_BOOK, LIQUIDATIONS, OPEN_INTEREST, PERPETUAL, SELL, SPOT, TICKER, TRADES, FILLED, UNFILLED
 from cryptofeed.feed import Feed
 from cryptofeed.symbols import Symbol
+from cryptofeed.exchanges.mixins.binance_rest import BinanceRestMixin
+from cryptofeed.types import Trade, Ticker, Candle, Liquidation, Funding, OrderBook
+
 
 LOG = logging.getLogger('feedhandler')
 
 
-class Binance(Feed):
+class Binance(Feed, BinanceRestMixin):
     id = BINANCE
     symbol_endpoint = 'https://api.binance.com/api/v3/exchangeInfo'
+    listen_key_endpoint = 'userDataStream'
     valid_depths = [5, 10, 20, 50, 100, 500, 1000, 5000]
     # m -> minutes; h -> hours; d -> days; w -> weeks; M -> months
     valid_candle_intervals = {'1m', '3m', '5m', '15m', '30m', '1h', '2h', '4h', '6h', '8h', '12h', '1d', '3d', '1w', '1M'}
@@ -32,7 +38,8 @@ class Binance(Feed):
         L2_BOOK: 'depth',
         TRADES: 'aggTrade',
         TICKER: 'bookTicker',
-        CANDLES: 'kline_'
+        CANDLES: 'kline_',
+        BALANCES: BALANCES
     }
     request_limit = 20
 
@@ -64,10 +71,8 @@ class Binance(Feed):
             info['instrument_type'][s.normalized] = stype
         return ret, info
 
-    def __init__(self, candle_interval='1m', candle_closed_only=False, depth_interval='100ms', concurrent_http=False, **kwargs):
+    def __init__(self, candle_closed_only=False, depth_interval='100ms', concurrent_http=False, **kwargs):
         """
-        candle_interval: str
-            time between candles updates ['1m', '3m', '5m', '15m', '30m', '1h', '2h', '4h', '6h', '8h', '12h', '1d', '3d', '1w', '1M']
         candle_closed_only: bool
             return only closed candles, i.e. no updates in between intervals.
         depth_interval: str
@@ -75,20 +80,19 @@ class Binance(Feed):
         concurrent_http: bool
             http requests will be made concurrently, if False requests will be made one at a time (affects L2_BOOK, OPEN_INTEREST).
         """
-        if candle_interval not in self.valid_candle_intervals:
-            raise ValueError(f"Candle interval must be one of {self.valid_candle_intervals}")
         if depth_interval is not None and depth_interval not in self.valid_depth_intervals:
             raise ValueError(f"Depth interval must be one of {self.valid_depth_intervals}")
 
         super().__init__({}, **kwargs)
         self.ws_endpoint = 'wss://stream.binance.com:9443'
-        self.rest_endpoint = 'https://www.binance.com/api/v1'
-        self.candle_interval = candle_interval
+        self.rest_endpoint = 'https://www.binance.com/api/v3'
         self.candle_closed_only = candle_closed_only
         self.depth_interval = depth_interval
         self.address = self._address()
         self.concurrent_http = concurrent_http
+        self.token = None
 
+        self._open_interest_cache = {}
         self._reset()
 
     def _address(self) -> Union[str, Dict]:
@@ -101,12 +105,23 @@ class Binance(Feed):
         The generic connect method supplied by Feed will take care of creating the
         correct connection objects from the addresses.
         """
-        address = self.ws_endpoint + '/stream?streams='
+        if self.requires_authentication:
+            listen_key = self._generate_token()
+            address = self.ws_endpoint + '/ws/' + listen_key
+        else:
+            address = self.ws_endpoint + '/stream?streams='
         subs = []
 
+        is_any_private = any(self.is_authenticated_channel(chan) for chan in self.subscription)
+        is_any_public = any(not self.is_authenticated_channel(chan) for chan in self.subscription)
+        if is_any_private and is_any_public:
+            raise ValueError("Private and public channels should be subscribed to in separate feeds")
+
         for chan in self.subscription:
             normalized_chan = self.exchange_channel_to_std(chan)
-            if self.exchange_channel_to_std(chan) == OPEN_INTEREST:
+            if normalized_chan == OPEN_INTEREST:
+                continue
+            if self.is_authenticated_channel(normalized_chan):
                 continue
 
             stream = chan
@@ -124,7 +139,7 @@ class Binance(Feed):
                     pair = pair.lower()
                 subs.append(f"{pair}@{stream}")
 
-        if len(subs) < 200:
+        if 0 < len(subs) < 200:
             return address + '/'.join(subs)
         else:
             def split_list(_list: list, n: int):
@@ -134,15 +149,33 @@ class Binance(Feed):
             return {chunk[0]: address + '/'.join(chunk) for chunk in split_list(subs, 200)}
 
     def _reset(self):
-        self.forced = defaultdict(bool)
         self._l2_book = {}
         self.last_update_id = {}
-        self.open_interest = {}
 
         if self.concurrent_http:
             # buffer 'depthUpdate' book msgs until snapshot is fetched
             self._book_buffer: Dict[str, deque[Tuple[dict, str, float]]] = {}
 
+    async def _refresh_token(self):
+        while True:
+            await sleep(30 * 60)
+            if self.token is None:
+                raise ValueError('There is no token to refresh')
+            payload = {'listenKey': self.token}
+            r = requests.put(f'{self.api}{self.listen_key_endpoint}?{urlencode(payload)}', headers={'X-MBX-APIKEY': self.key_id})
+            r.raise_for_status()
+
+    def _generate_token(self) -> str:
+        url = f'{self.api}{self.listen_key_endpoint}'
+        r = requests.post(url, headers={'X-MBX-APIKEY': self.key_id})
+        r.raise_for_status()
+        response = r.json()
+        if 'listenKey' in response:
+            self.token = response['listenKey']
+            return self.token
+        else:
+            raise ValueError(f'Unable to retrieve listenKey token from {url}')
+
     async def _trade(self, msg: dict, timestamp: float):
         """
         {
@@ -159,16 +192,15 @@ class Binance(Feed):
             "M": true         // Ignore
         }
         """
-        price = Decimal(msg['p'])
-        amount = Decimal(msg['q'])
-        await self.callback(TRADES, feed=self.id,
-                            order_id=msg['a'],
-                            symbol=self.exchange_symbol_to_std_symbol(msg['s']),
-                            side=SELL if msg['m'] else BUY,
-                            amount=amount,
-                            price=price,
-                            timestamp=self.timestamp_normalize(msg['E']),
-                            receipt_timestamp=timestamp)
+        t = Trade(self.id,
+                  self.exchange_symbol_to_std_symbol(msg['s']),
+                  SELL if msg['m'] else BUY,
+                  Decimal(msg['q']),
+                  Decimal(msg['p']),
+                  self.timestamp_normalize(msg['E']),
+                  id=str(msg['a']),
+                  raw=msg)
+        await self.callback(TRADES, t, timestamp)
 
     async def _ticker(self, msg: dict, timestamp: float):
         """
@@ -191,12 +223,8 @@ class Binance(Feed):
         else:
             ts = timestamp
 
-        await self.callback(TICKER, feed=self.id,
-                            symbol=pair,
-                            bid=bid,
-                            ask=ask,
-                            timestamp=ts,
-                            receipt_timestamp=timestamp)
+        t = Ticker(self.id, pair, bid, ask, ts, raw=msg)
+        await self.callback(TICKER, t, timestamp)
 
     async def _liquidations(self, msg: dict, timestamp: float):
         """
@@ -219,42 +247,37 @@ class Binance(Feed):
         }
         """
         pair = self.exchange_symbol_to_std_symbol(msg['o']['s'])
-        await self.callback(LIQUIDATIONS,
-                            feed=self.id,
-                            symbol=pair,
-                            side=msg['o']['S'],
-                            leaves_qty=Decimal(msg['o']['q']),
-                            price=Decimal(msg['o']['p']),
-                            order_id=None,
-                            status=FILLED if msg['o']['X'] == 'FILLED' else UNFILLED,
-                            timestamp=self.timestamp_normalize(msg['E']),
-                            receipt_timestamp=timestamp)
+        liq = Liquidation(self.id,
+                          pair,
+                          SELL if msg['o']['S'] == 'SELL' else BUY,
+                          Decimal(msg['o']['q']),
+                          Decimal(msg['o']['p']),
+                          None,
+                          FILLED if msg['o']['X'] == 'FILLED' else UNFILLED,
+                          self.timestamp_normalize(msg['E']),
+                          raw=msg)
+        await self.callback(LIQUIDATIONS, liq, receipt_timestamp=timestamp)
 
-    def _check_update_id(self, std_pair: str, msg: dict) -> Tuple[bool, bool, bool]:
+    def _check_update_id(self, std_pair: str, msg: dict) -> bool:
         """
-        'current_match' is True if the msg's update_id matches snapshot's update_id in self.last_update_id.
-        Messages will be queued (or buffered if concurrent_http is True) while fetching for snapshot, and we can return a book_callback using this msg's data instead of waiting for the next update.
+        Messages will be queued (or buffered if concurrent_http is True) while fetching for snapshot
+        and we can return a book_callback using this msg's data instead of waiting for the next update.
         """
-        skip_update = False
-        forced = not self.forced[std_pair]
-        current_match = msg['u'] == self.last_update_id[std_pair]
-
-        if forced and msg['u'] <= self.last_update_id[std_pair]:
-            skip_update = True
-        elif forced and msg['U'] <= self.last_update_id[std_pair] + 1 <= msg['u']:
+        if self._l2_book[std_pair].delta is None and msg['u'] <= self.last_update_id[std_pair]:
+            return True
+        elif msg['U'] <= self.last_update_id[std_pair] + 1 <= msg['u']:
             self.last_update_id[std_pair] = msg['u']
-            self.forced[std_pair] = True
-        elif not forced and self.last_update_id[std_pair] + 1 == msg['U']:
+            return False
+        elif self.last_update_id[std_pair] + 1 == msg['U']:
             self.last_update_id[std_pair] = msg['u']
+            return False
         else:
             self._reset()
             LOG.warning("%s: Missing book update detected, resetting book", self.id)
-            skip_update = True
-
-        return skip_update, forced, current_match
+            return True
 
     async def _snapshot(self, pair: str) -> None:
-        max_depth = self.max_depth if self.max_depth else 1000
+        max_depth = self.max_depth if self.max_depth else self.valid_depths[-1]
         if max_depth not in self.valid_depths:
             for d in self.valid_depths:
                 if d > max_depth:
@@ -267,42 +290,34 @@ class Binance(Feed):
 
         std_pair = self.exchange_symbol_to_std_symbol(pair)
         self.last_update_id[std_pair] = resp['lastUpdateId']
-        self._l2_book[std_pair] = {BID: sd(), ASK: sd()}
-        for s, side in (('bids', BID), ('asks', ASK)):
-            for update in resp[s]:
-                price = Decimal(update[0])
-                amount = Decimal(update[1])
-                self._l2_book[std_pair][side][price] = amount
+        self._l2_book[std_pair] = OrderBook(self.id, std_pair, max_depth=self.max_depth, bids={Decimal(u[0]): Decimal(u[1]) for u in resp['bids']}, asks={Decimal(u[0]): Decimal(u[1]) for u in resp['asks']})
+        ts = self.timestamp_normalize(resp['E']) if 'E' in resp else None
+        await self.book_callback(L2_BOOK, self._l2_book[std_pair], time.time(), raw=resp, timestamp=ts, sequence_number=resp['lastUpdateId'])
 
     async def _handle_book_msg(self, msg: dict, pair: str, timestamp: float):
         """
         Processes 'depthUpdate' update book msg. This method should only be called if pair's l2_book exists.
         """
         std_pair = self.exchange_symbol_to_std_symbol(pair)
-        skip_update, forced, current_match = self._check_update_id(std_pair, msg)
-        if current_match:
-            # Current msg.final_update_id == self.last_update_id[pair] which is the snapshot's update id
-            return await self.book_callback(self._l2_book[std_pair], L2_BOOK, std_pair, forced, self._l2_book[std_pair], self.timestamp_normalize(msg['E']), timestamp)
-
+        skip_update = self._check_update_id(std_pair, msg)
         if skip_update:
             return
 
         delta = {BID: [], ASK: []}
-        ts = msg['E']
-
         for s, side in (('b', BID), ('a', ASK)):
             for update in msg[s]:
                 price = Decimal(update[0])
                 amount = Decimal(update[1])
 
                 if amount == 0:
-                    if price in self._l2_book[std_pair][side]:
-                        del self._l2_book[std_pair][side][price]
+                    if price in self._l2_book[std_pair].book[side]:
+                        del self._l2_book[std_pair].book[side][price]
                         delta[side].append((price, amount))
                 else:
-                    self._l2_book[std_pair][side][price] = amount
+                    self._l2_book[std_pair].book[side][price] = amount
                     delta[side].append((price, amount))
-        await self.book_callback(self._l2_book[std_pair], L2_BOOK, std_pair, forced, delta, self.timestamp_normalize(ts), timestamp)
+
+        await self.book_callback(L2_BOOK, self._l2_book[std_pair], timestamp, timestamp=self.timestamp_normalize(msg['E']), raw=msg, delta=delta, sequence_number=self.last_update_id[std_pair])
 
     async def _book(self, msg: dict, pair: str, timestamp: float) -> None:
         """
@@ -367,16 +382,28 @@ class Binance(Feed):
             "r": "0.00030000",       // Funding rate
             "T": 1562306400000       // Next funding time
         }
+
+        BinanceFutures
+        {
+            "e": "markPriceUpdate",     // Event type
+            "E": 1562305380000,         // Event time
+            "s": "BTCUSDT",             // Symbol
+            "p": "11185.87786614",      // Mark price
+            "i": "11784.62659091"       // Index price
+            "P": "11784.25641265",      // Estimated Settle Price, only useful in the last hour before the settlement starts
+            "r": "0.00030000",          // Funding rate
+            "T": 1562306400000          // Next funding time
+        }
         """
-        await self.callback(FUNDING,
-                            feed=self.id,
-                            symbol=self.exchange_symbol_to_std_symbol(msg['s']),
-                            timestamp=self.timestamp_normalize(msg['E']),
-                            receipt_timestamp=timestamp,
-                            mark_price=msg['p'],
-                            rate=msg['r'],
-                            next_funding_time=self.timestamp_normalize(msg['T']),
-                            )
+        f = Funding(self.id,
+                    self.exchange_symbol_to_std_symbol(msg['s']),
+                    Decimal(msg['p']),
+                    Decimal(msg['r']),
+                    self.timestamp_normalize(msg['T']),
+                    self.timestamp_normalize(msg['E']),
+                    predicted_rate=Decimal(msg['P']) if 'P' in msg and msg['P'] is not None else None,
+                    raw=msg)
+        await self.callback(FUNDING, f, timestamp)
 
     async def _candle(self, msg: dict, timestamp: float):
         """
@@ -407,26 +434,54 @@ class Binance(Feed):
         """
         if self.candle_closed_only and not msg['k']['x']:
             return
+        c = Candle(self.id,
+                   self.exchange_symbol_to_std_symbol(msg['s']),
+                   msg['k']['t'] / 1000,
+                   msg['k']['T'] / 1000,
+                   msg['k']['i'],
+                   msg['k']['n'],
+                   Decimal(msg['k']['o']),
+                   Decimal(msg['k']['c']),
+                   Decimal(msg['k']['h']),
+                   Decimal(msg['k']['l']),
+                   Decimal(msg['k']['v']),
+                   msg['k']['x'],
+                   self.timestamp_normalize(msg['E']),
+                   raw=msg)
+        await self.callback(CANDLES, c, timestamp)
 
-        await self.callback(CANDLES,
-                            feed=self.id,
-                            symbol=self.exchange_symbol_to_std_symbol(msg['s']),
-                            timestamp=self.timestamp_normalize(msg['E']),
-                            receipt_timestamp=timestamp,
-                            start=msg['k']['t'] / 1000,
-                            stop=msg['k']['T'] / 1000,
-                            interval=msg['k']['i'],
-                            trades=msg['k']['n'],
-                            open_price=Decimal(msg['k']['o']),
-                            close_price=Decimal(msg['k']['c']),
-                            high_price=Decimal(msg['k']['h']),
-                            low_price=Decimal(msg['k']['l']),
-                            volume=Decimal(msg['k']['v']),
-                            closed=msg['k']['x'])
+    async def _account_update(self, msg: dict, timestamp: float):
+        """
+        {
+            "e": "outboundAccountPosition", //Event type
+            "E": 1564034571105,             //Event Time
+            "u": 1564034571073,             //Time of last account update
+            "B": [                          //Balances Array
+                {
+                "a": "ETH",                 //Asset
+                "f": "10000.000000",        //Free
+                "l": "0.000000"             //Locked
+                }
+            ]
+        }
+        """
+        for balance in msg['B']:
+            await self.callback(BALANCES,
+                                feed=self.id,
+                                symbol=balance['a'],
+                                timestamp=self.timestamp_normalize(msg['E']),
+                                receipt_timestamp=timestamp,
+                                wallet_balance=Decimal(balance['f']))
 
     async def message_handler(self, msg: str, conn, timestamp: float):
         msg = json.loads(msg, parse_float=Decimal)
 
+        # Handle account updates from User Data Stream
+        if self.requires_authentication:
+            msg_type = msg['e']
+            if msg_type == 'outboundAccountPosition':
+                await self._account_update(msg, timestamp)
+            return
         # Combined stream events are wrapped as follows: {"stream":"<streamName>","data":<rawPayload>}
         # streamName is of format <symbol>@<channel>
         pair, _ = msg['stream'].split('@', 1)
@@ -454,5 +509,9 @@ class Binance(Feed):
         # Binance does not have a separate subscribe message, the
         # subscription information is included in the
         # connection endpoint
-        if not isinstance(conn, HTTPPoll):
+        if isinstance(conn, (HTTPPoll, HTTPConcurrentPoll)):
+            self._open_interest_cache = {}
+        else:
             self._reset()
+        if self.requires_authentication:
+            create_task(self._refresh_token())
